@@ -86,7 +86,7 @@
 //!
 //! Guard nodes were first proposes (as "helper nodes") in "Defending
 //! Anonymous Communications Against Passive Logging Attacks" by
-//! Matther Wright, Micah Adler, Brian N. Levine, and Clay Shields in
+//! Matthew Wright, Micah Adler, Brian N. Levine, and Clay Shields in
 //! the Proceedings of the 2003 IEEE Symposium on Security and
 //! Privacy.  (See <https://www.freehaven.net/anonbib/#wright03>)
 //!
@@ -97,6 +97,7 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
+#![deny(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
@@ -118,6 +119,7 @@
 #![warn(clippy::option_option)]
 #![warn(clippy::rc_buffer)]
 #![deny(clippy::ref_option_ref)]
+#![warn(clippy::semicolon_if_nothing_returned)]
 #![warn(clippy::trait_duplication_in_bounds)]
 #![deny(clippy::unnecessary_wraps)]
 #![warn(clippy::unseparated_literal_suffix)]
@@ -132,7 +134,7 @@
 use futures::channel::mpsc;
 use futures::task::{SpawnError, SpawnExt};
 use serde::{Deserialize, Serialize};
-use std::collections::HashMap;
+use std::collections::{HashMap, HashSet};
 use std::convert::{TryFrom, TryInto};
 use std::net::SocketAddr;
 use std::sync::{Arc, Mutex};
@@ -178,16 +180,11 @@ pub struct GuardMgr<R: Runtime> {
 /// This would just be a [`GuardMgr`], except that it needs to sit inside
 /// a `Mutex` and get accessed by daemon tasks.
 struct GuardMgrInner {
-    /// Last time when we've taken note of activity that means we're
-    /// online.
+    /// Last time when marked all of our primary guards as retriable.
     ///
-    /// We use this timestamp when a low-priority guard is discovered
-    /// to be online.  If it happens when we've been _offline_ for a
-    /// while, then we retry our primary guards before using the low
-    /// priority guard, so that we can be sure that they're really
-    /// down (and didn't just _seem_ to be down because we were
-    /// offline).
-    last_time_on_internet: Option<Instant>,
+    /// We keep track of this time so that we can rate-limit
+    /// these attempts.
+    last_primary_retry_time: Instant,
 
     /// The currently active [`GuardSet`] object.
     ///
@@ -208,13 +205,13 @@ struct GuardMgrInner {
     /// A mpsc channel, used to tell the task running in
     /// [`daemon::report_status_events`] about a new event to monitor.
     ///
-    /// This uses an `UnboundedSener` so that we don't have to await
+    /// This uses an `UnboundedSender` so that we don't have to await
     /// while sending the message, which in turn allows the GuardMgr
     /// API to be simpler.  The risk, however, is that there's no
     /// backpressure in the event that the task running
     /// [`daemon::report_status_events`] fails to read from this
     /// channel.
-    ctrl: mpsc::UnboundedSender<daemon::MsgResult>,
+    ctrl: mpsc::UnboundedSender<daemon::Msg>,
 
     /// Information about guards that we've given out, but where we have
     /// not yet heard whether the guard was successful.
@@ -255,7 +252,7 @@ impl<R: Runtime> GuardMgr<R> {
         let active_guards = default_storage.load()?.unwrap_or_else(GuardSet::new);
         let inner = Arc::new(Mutex::new(GuardMgrInner {
             active_guards,
-            last_time_on_internet: None,
+            last_primary_retry_time: runtime.now(),
             params: GuardParams::default(),
             ctrl,
             pending: HashMap::new(),
@@ -309,6 +306,12 @@ impl<R: Runtime> GuardMgr<R> {
         Ok(())
     }
 
+    /// Return true if `netdir` has enough information to safely become our new netdir.
+    pub fn netdir_is_sufficient(&self, netdir: &NetDir) -> bool {
+        let mut inner = self.inner.lock().expect("Poisoned lock");
+        inner.active_guards.missing_primary_microdescriptors(netdir) == 0
+    }
+
     /// Update the state of this [`GuardMgr`] based on a new or modified
     /// [`NetDir`] object.
     ///
@@ -322,28 +325,9 @@ impl<R: Runtime> GuardMgr<R> {
         let now = self.runtime.wallclock();
 
         let mut inner = self.inner.lock().expect("Poisoned lock");
-
         inner.update(now, Some(netdir));
     }
 
-    /// Record that some internet activity has happened that tells us
-    /// we're online.
-    ///
-    /// We can use this information to determine when we should retry
-    /// our primary guards on the basis of having been down for a long
-    /// time.
-    ///
-    /// # Limitations
-    ///
-    /// We should really call this every time we read a cell, but that
-    /// isn't efficient or practical.  We'll probably have to refactor
-    /// things somehow. (TODO)
-    pub fn note_internet_activity(&self) {
-        let now = self.runtime.now();
-        let mut inner = self.inner.lock().expect("Poisoned lock");
-        inner.last_time_on_internet = Some(now);
-    }
-
     /// Replace the current [`GuardFilter`] used by this `GuardMgr`.
     ///
     /// (Since there is only one kind of filter right now, there's no
@@ -370,7 +354,14 @@ impl<R: Runtime> GuardMgr<R> {
 
         // TODO: Once we support nontrivial filters, we might have to
         // swap out "active_guards" depending on which set it is.
-        // TODO: Warn if the filter is waaaay to small according to guard params.
+
+        if frac_permitted < inner.params.extreme_threshold {
+            warn!(
+                "The number of guards permitted is smaller than the guard param minimum of {}%.",
+                inner.params.extreme_threshold * 100.0,
+            );
+        }
+
         info!(
             ?filter,
             restrictive = restrictive_filter,
@@ -401,7 +392,8 @@ impl<R: Runtime> GuardMgr<R> {
     /// the [`NetDir`] most recently passed to [`GuardMgr::update_network`].
     /// That's _usually_ what you'd want, but when we're trying to
     /// bootstrap we might want to use _all_ guards as possible
-    /// directory caches.  That's not implemented yet.
+    /// directory caches.  That's not implemented yet. (See ticket
+    /// [#220](https://gitlab.torproject.org/tpo/core/arti/-/issues/220)).
     ///
     /// This function only looks at netdir when all of the known
     /// guards are down; to force an update, use [`GuardMgr::update_network`].
@@ -435,23 +427,27 @@ impl<R: Runtime> GuardMgr<R> {
             (u, Some(snd))
         };
         let request_id = pending::RequestId::next();
-        let (monitor, rcv) = GuardMonitor::new(request_id);
+        let ctrl = inner.ctrl.clone();
+        let monitor = GuardMonitor::new(request_id, ctrl);
+
+        // Note that the network can be down even if all the primary guards
+        // are not yet marked as unreachable.  But according to guard-spec we
+        // don't want to acknowledge the net as down before that point, since
+        // we don't mark all the primary guards as retriable unless
+        // we've been forced to non-primary guards.
+        let net_has_been_down = inner.active_guards.all_primary_guards_are_unreachable()
+            && tor_proto::time_since_last_incoming_traffic() >= inner.params.internet_down_timeout;
 
         let pending_request =
-            pending::PendingRequest::new(guard_id.clone(), usage, usable_sender, now);
+            pending::PendingRequest::new(guard_id.clone(), usage, usable_sender, net_has_been_down);
         inner.pending.insert(request_id, pending_request);
 
         inner.active_guards.record_attempt(&guard_id, now);
 
-        inner
-            .ctrl
-            .unbounded_send(Ok(daemon::Msg::Observe(rcv)))
-            .expect("Guard observer task exited prematurely");
-
         Ok((guard, monitor, usable))
     }
 
-    /// Ensure that the message queue is flushed before proceding to
+    /// Ensure that the message queue is flushed before proceeding to
     /// the next step.  Used for testing.
     #[cfg(test)]
     async fn flush_msg_queue(&self) {
@@ -461,8 +457,8 @@ impl<R: Runtime> GuardMgr<R> {
             let inner = self.inner.lock().expect("Poisoned lock");
             inner
                 .ctrl
-                .unbounded_send(Ok(pingmsg))
-                .expect("Guard observer task exited permaturely.");
+                .unbounded_send(pingmsg)
+                .expect("Guard observer task exited prematurely.");
         }
         let _ = rcv.await;
     }
@@ -487,6 +483,11 @@ impl GuardMgrInner {
         self.active_guards.expire_old_guards(&self.params, now);
 
         if let Some(netdir) = netdir {
+            if self.active_guards.missing_primary_microdescriptors(netdir) > 0 {
+                // We are missing primary guard descriptors, so we shouldn't update our guard
+                // status.
+                return;
+            }
             self.active_guards.update_status_from_netdir(netdir);
             loop {
                 let added_any =
@@ -509,6 +510,25 @@ impl GuardMgrInner {
         self.update(now, None);
     }
 
+    /// Mark all of our primary guards as retriable, if we haven't done
+    /// so since long enough before `now`.
+    ///
+    /// We want to call this function whenever a guard attempt succeeds,
+    /// if the internet seemed to be down when the guard attempt was
+    /// first launched.
+    fn maybe_retry_primary_guards(&mut self, now: Instant) {
+        // We don't actually want to mark our primary guards as
+        // retriable more than once per internet_down_timeout: after
+        // the first time, we would just be noticing the same "coming
+        // back online" event more than once.
+        let interval = self.params.internet_down_timeout;
+        if self.last_primary_retry_time + interval <= now {
+            debug!("Successfully reached a guard after a while off the internet; marking all primary guards retriable.");
+            self.active_guards.mark_primary_guards_retriable();
+            self.last_primary_retry_time = now;
+        }
+    }
+
     /// Called when the circuit manager reports (via [`GuardMonitor`]) that
     /// a guard succeeded or failed.
     ///
@@ -526,23 +546,13 @@ impl GuardMgrInner {
             trace!(?guard_id, ?status, "Received report of guard status");
             match status {
                 GuardStatus::Success => {
-                    let now = runtime.now();
-                    // If we've been gone too long without any net activitity,
-                    // and now we're seeing a circuit succeed,
-                    // tell the primary guards that they might be retriable.
-                    if let Some(last_time) = self.last_time_on_internet {
-                        let dur = now.saturating_duration_since(last_time);
-                        // TODO: we should use the actual timeout for this, but
-                        // we can't do it yet, since we don't have anything reliable
-                        // calling note_internet_activity.
-                        // let timeout = self.params.internet_down_timeout;
-                        let timeout = Duration::from_secs(7200); // (Fake timeout)
-                        if dur >= timeout {
-                            debug!("Retrying primary guards");
-                            self.active_guards.mark_primary_guards_retriable();
-                        }
+                    // If we had gone too long without any net activity when we
+                    // gave out this guard, and now we're seeing a circuit
+                    // succeed, tell the primary guards that they might be
+                    // retriable.
+                    if pending.net_has_been_down() {
+                        self.maybe_retry_primary_guards(runtime.now());
                     }
-                    self.last_time_on_internet = Some(now);
 
                     // The guard succeeded.  Tell the GuardSet.
                     self.active_guards
@@ -609,7 +619,7 @@ impl GuardMgrInner {
     /// expire them and tell the circuit manager that their circuits
     /// are unusable.
     fn expire_and_answer_pending_requests(&mut self, now: Instant) {
-        // TODO: Use Vec::drain_filter when/if it's stable.
+        // TODO: Use Vec::drain_filter or Vec::retain_mut when/if it's stable.
         use retain_mut::RetainMut;
 
         // A bit ugly: we use a separate Vec here to avoid borrowing issues,
@@ -617,7 +627,7 @@ impl GuardMgrInner {
         let mut waiting = Vec::new();
         std::mem::swap(&mut waiting, &mut self.waiting);
 
-        waiting.retain_mut(|pending| {
+        RetainMut::retain_mut(&mut waiting, |pending| {
             let expired = pending
                 .waiting_since()
                 .and_then(|w| now.checked_duration_since(w))
@@ -732,13 +742,15 @@ struct GuardParams {
     /// After how much time without successful activity does a
     /// successful circuit indicate that we should retry our primary
     /// guards?
-    #[allow(dead_code)] // XXXX not yet implemented.
     internet_down_timeout: Duration,
     /// What fraction of the guards can be can be filtered out before we
     /// decide that our filter is "very restrictive"?
     ///
     /// (Not fully implemented yet.)
     filter_threshold: f64,
+    /// What fraction of the guards determine that our filter is "very
+    /// restrictive"?
+    extreme_threshold: f64,
 }
 
 impl Default for GuardParams {
@@ -758,6 +770,7 @@ impl Default for GuardParams {
             np_idle_timeout: Duration::from_secs(600),
             internet_down_timeout: Duration::from_secs(600),
             filter_threshold: 0.2,
+            extreme_threshold: 0.01,
         }
     }
 }
@@ -779,6 +792,7 @@ impl TryFrom<&NetParameters> for GuardParams {
             np_idle_timeout: p.guard_nonprimary_idle_timeout.try_into()?,
             internet_down_timeout: p.guard_internet_likely_down.try_into()?,
             filter_threshold: p.guard_meaningful_restriction.as_fraction(),
+            extreme_threshold: p.guard_extreme_restriction.as_fraction(),
         })
     }
 }
@@ -874,15 +888,14 @@ impl Default for GuardUsageKind {
 ///
 /// Used as an argument to [`GuardMgr::select_guard`].
 #[derive(Clone, Debug, Default, derive_builder::Builder)]
+#[builder(build_fn(error = "tor_config::ConfigBuildError"))]
 pub struct GuardUsage {
     /// The purpose for which this guard will be used.
     #[builder(default)]
     kind: GuardUsageKind,
-    /// An optional restriction on which guard may be used.
-    ///
-    /// (Eventually, multiple restrictions may be supported.)
-    #[builder(default, setter(strip_option))]
-    restriction: Option<GuardRestriction>,
+    /// A list of restrictions on which guard may be used.
+    #[builder(default)]
+    restrictions: Vec<GuardRestriction>,
 }
 
 impl GuardUsageBuilder {
@@ -890,6 +903,14 @@ impl GuardUsageBuilder {
     pub fn new() -> Self {
         Self::default()
     }
+
+    /// Add `restriction` to the list of restrictions on this guard usage.
+    pub fn push_restriction(&mut self, restriction: GuardRestriction) -> &mut Self {
+        self.restrictions
+            .get_or_insert_with(Vec::new)
+            .push(restriction);
+        self
+    }
 }
 
 /// A restriction that applies to a single request for a guard.
@@ -904,6 +925,8 @@ impl GuardUsageBuilder {
 pub enum GuardRestriction {
     /// Don't pick a guard with the provided Ed25519 identity.
     AvoidId(pk::ed25519::Ed25519Identity),
+    /// Don't pick a guard with any of the provided Ed25519 identities.
+    AvoidAllIds(HashSet<pk::ed25519::Ed25519Identity>),
 }
 
 /// An error caused while creating or updating a guard manager.
@@ -987,12 +1010,14 @@ mod test {
             let usage = GuardUsage::default();
             let (id2, _mon, _usable) = guardmgr2.select_guard(usage, Some(&netdir)).unwrap();
             assert_eq!(id2, id);
-        })
+        });
     }
 
     #[test]
     fn simple_waiting() {
-        // XXXX This test fails in rare cases; I suspect a race condition somewhere.
+        // TODO(nickm): This test fails in rare cases; I suspect a
+        // race condition somewhere.
+        //
         // I've doubled up on the queue flushing in order to try to make the
         // race less likely, but we should investigate.
         test_with_all_runtimes!(|rt| async move {
@@ -1031,7 +1056,7 @@ mod test {
             );
 
             assert_eq!((u3, u4), (false, true));
-        })
+        });
     }
 
     #[test]
@@ -1045,6 +1070,6 @@ mod test {
             let (guard, _mon, _usable) = guardmgr.select_guard(u, Some(&netdir)).unwrap();
             // Make sure that the filter worked.
             assert_eq!(guard.id().rsa.as_bytes()[0] % 4, 0);
-        })
+        });
     }
 }
