@@ -173,7 +173,7 @@ impl GuardSet {
 
     /// Copy non-persistent status from every guard shared with `other`.
     pub(crate) fn copy_status_from(&mut self, other: &GuardSet) {
-        for (id, guard) in self.guards.iter_mut() {
+        for (id, guard) in &mut self.guards {
             if let Some(other_guard) = other.get(id) {
                 guard.copy_status_from(other_guard);
             }
@@ -384,6 +384,18 @@ impl GuardSet {
         self.primary_guards_invalidated = true;
     }
 
+    /// Return the number of our primary guards are missing their
+    /// microdescriptors in `dir`.
+    pub(crate) fn missing_primary_microdescriptors(&mut self, dir: &NetDir) -> usize {
+        self.primary
+            .iter()
+            .filter(|id| {
+                let g = self.guards.get(id).expect("Inconsistent guard state");
+                g.listed_in(dir).is_none()
+            })
+            .count()
+    }
+
     /// Update the status of every guard  in this sample from a network
     /// directory.
     pub(crate) fn update_status_from_netdir(&mut self, dir: &NetDir) {
@@ -438,7 +450,7 @@ impl GuardSet {
         }
 
         // Clear exploratory_circ_pending for all primary guards.
-        for id in self.primary.iter() {
+        for id in &self.primary {
             if let Some(guard) = self.guards.get_mut(id) {
                 guard.note_exploratory_circ(false);
             }
@@ -519,16 +531,25 @@ impl GuardSet {
 
     /// Mark every `Unreachable` primary guard as `Unknown`.
     pub(crate) fn mark_primary_guards_retriable(&mut self) {
-        for id in self.primary.iter() {
+        for id in &self.primary {
             if let Some(g) = self.guards.get_mut(id) {
                 g.mark_retriable();
             }
         }
     }
 
+    /// Return true if all of our primary guards are currently marked
+    /// unreachable.
+    pub(crate) fn all_primary_guards_are_unreachable(&mut self) -> bool {
+        self.primary
+            .iter()
+            .flat_map(|id| self.guards.get(id))
+            .all(|g| g.reachable() == Reachable::Unreachable)
+    }
+
     /// Mark every `Unreachable` guard as `Unknown`.
     pub(crate) fn mark_all_guards_retriable(&mut self) {
-        for (_, g) in self.guards.iter_mut() {
+        for g in self.guards.values_mut() {
             g.mark_retriable();
         }
     }
@@ -580,7 +601,7 @@ impl GuardSet {
     /// just been abandoned, without learning whether it succeeded or failed.
     pub(crate) fn record_attempt_abandoned(&mut self, guard_id: &GuardId) {
         if let Some(guard) = self.guards.get_mut(guard_id) {
-            guard.note_exploratory_circ(false)
+            guard.note_exploratory_circ(false);
         }
     }
 
@@ -784,7 +805,7 @@ mod test {
             guards.assert_consistency();
 
             // make sure all the guards are okay.
-            for (g, guard) in guards.guards.iter() {
+            for (g, guard) in &guards.guards {
                 let relay = g.get_relay(&netdir).unwrap();
                 assert!(relay.is_flagged_guard());
                 assert!(guards.contains_relay(&relay));
@@ -833,7 +854,7 @@ mod test {
             guards.guards.keys().collect::<HashSet<_>>(),
             guards2.guards.keys().collect::<HashSet<_>>()
         );
-        for (k, g) in guards.guards.iter() {
+        for (k, g) in &guards.guards {
             let g2 = guards2.guards.get(k).unwrap();
             assert_eq!(format!("{:?}", g), format!("{:?}", g2));
         }
@@ -904,7 +925,7 @@ mod test {
         // Pick a guard and mark it as confirmed.
         let id1 = guards.sample[0].clone();
         guards.record_success(&id1, &params, t1);
-        assert_eq!(&guards.confirmed, &[id1.clone()]);
+        assert_eq!(&guards.confirmed, &[id1]);
 
         let one_day = Duration::from_secs(86400);
         guards.expire_old_guards(&params, t1 + one_day * 30);
@@ -1078,4 +1099,75 @@ mod test {
         guards.select_primary_guards(&params);
         assert_eq!(guards.sample.len(), 10);
     }
+
+    #[test]
+    fn retry_primary() {
+        let netdir = netdir();
+        let params = GuardParams {
+            min_filtered_sample_size: 5,
+            n_primary: 2,
+            max_sample_bw_fraction: 1.0,
+            ..GuardParams::default()
+        };
+        let usage = crate::GuardUsageBuilder::default().build().unwrap();
+
+        let mut guards = GuardSet::new();
+
+        guards.extend_sample_as_needed(SystemTime::now(), &params, &netdir);
+        guards.select_primary_guards(&params);
+
+        assert_eq!(guards.primary.len(), 2);
+        assert!(!guards.all_primary_guards_are_unreachable());
+
+        // Let one primary guard fail.
+        let (kind, p_id1) = guards.pick_guard(&usage, &params).unwrap();
+        assert_eq!(kind, ListKind::Primary);
+        guards.record_failure(&p_id1, Instant::now());
+        assert!(!guards.all_primary_guards_are_unreachable());
+
+        // Now let the other one fail.
+        let (kind, p_id2) = guards.pick_guard(&usage, &params).unwrap();
+        assert_eq!(kind, ListKind::Primary);
+        guards.record_failure(&p_id2, Instant::now());
+        assert!(guards.all_primary_guards_are_unreachable());
+
+        // Now mark the guards retriable.
+        guards.mark_primary_guards_retriable();
+        assert!(!guards.all_primary_guards_are_unreachable());
+        let (kind, p_id3) = guards.pick_guard(&usage, &params).unwrap();
+        assert_eq!(kind, ListKind::Primary);
+        assert_eq!(p_id3, p_id1);
+    }
+
+    #[test]
+    fn count_missing_mds() {
+        let netdir = netdir();
+        let params = GuardParams {
+            min_filtered_sample_size: 5,
+            n_primary: 2,
+            max_sample_bw_fraction: 1.0,
+            ..GuardParams::default()
+        };
+        let usage = crate::GuardUsageBuilder::default().build().unwrap();
+        let mut guards = GuardSet::new();
+        guards.extend_sample_as_needed(SystemTime::now(), &params, &netdir);
+        guards.select_primary_guards(&params);
+        assert_eq!(guards.primary.len(), 2);
+
+        let (_kind, p_id1) = guards.pick_guard(&usage, &params).unwrap();
+        guards.record_success(&p_id1, &params, SystemTime::now());
+        assert_eq!(guards.missing_primary_microdescriptors(&netdir), 0);
+
+        use tor_netdir::testnet;
+        let netdir2 = testnet::construct_custom_netdir(|idx, bld| {
+            if idx == p_id1.ed25519.as_bytes()[0] as usize {
+                bld.omit_md = true;
+            }
+        })
+        .unwrap()
+        .unwrap_if_sufficient()
+        .unwrap();
+
+        assert_eq!(guards.missing_primary_microdescriptors(&netdir2), 1);
+    }
 }
